from .cell import CELL_CLASS_NAME


class Pype(dict):
    """A dict-like collection of Cells"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__.update({key: value})
        if type(self[key]).__name__ == CELL_CLASS_NAME:
            self[key].name = key

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    @property
    def cells(self):
        """Returns dict {cell_name: cell_instance} for every cell in this pype"""
        return {
            k: v for k, v in self.items() if type(v).__name__ == CELL_CLASS_NAME
        }

    def recalc_all(self):
        """Recalc all cells in this pype"""
        for cell in self.cells.values():
            cell.recalc()

    def dump_values(self):
        """Return all recalculated cell values in a list of dicts [{'name': ..., 'value': ...}]"""
        self.recalc_all()
        return [
            {
                "name": key,
                "value": cell._value.copy()
                if "pandas" in str(type(cell._value))
                else cell._value,
            }
            for key, cell in self.cells.items()
        ]

    def load_values(self, name_value_list):
        """Restore cell values from a list of dicts [{'name': ..., 'value': ...}]"""
        for item in name_value_list:
            self[item["name"]]._value = item["value"]
            self[item["name"]]._dirty = False

    ## in order to load/dump entire Pypes, including dependencies and recalc functions, consider dill:
    ## given a 'pype' collection with cells:
    #
    # import dill
    #
    # with open('pype_p.dill', 'wb') as outp:
    #    dill.dump(pype, outp)
    # with open('pype_p.dill', 'rb') as inp:
    #    pype_2 = dill.load(inp)
    #
    ## note 1. security concerns: code injectable by anyone with acess to dill file
    ##      2. fragile to lib versions: do not use as a structural persistance mechanism



from .helpers import deep_eq

DEBUG = False


class Cell:
    """Cell class to support reactive dataflow-style dynamic evaluation of DataFrames"""

    def __init__(
        self,
        sources: list = None,
        recalc: callable = None,
        on_change: callable = None,
        lazy: bool = True,
    ):
        """
        Args:
            sources: List of cell instances used to recalc this cell
            recalc: Recalc function; the number of (position-based) arguments to this function must equal len(sources)
            on_change: External function to be called when value changes; cell.value is passed as argument (implies lazy=False)
            lazy: If True (default): do not recalc cell on invalidation but wait for (a) value getter or (b) explicit recalc call
        """
        self.name = ""  # optional, set by Pype
        self.sources = sources or []
        self.recalc_handler = recalc or (lambda *args: None)
        self.on_change_handler = on_change or None
        self.lazy = lazy and not (on_change)
        # init internals:
        self._previous = None
        self._dirty = False
        self._value = None
        self._sinks = []
        for cell in self.sources:
            cell._sinks.append(self)
        self.invalidate()

    def _invalidate_dependents(self):
        for s in self._sinks:
            s.invalidate()

    def invalidate(self):
        if DEBUG and self.name:
            print("%s invalidated" % self.name)
        if self.lazy:
            self._dirty = True
            self._invalidate_dependents()
        else:
            self.recalc()

    @property  # @value.getter
    def value(self):
        if self._dirty:
            self.recalc()
        return self._value

    @value.setter
    def value(self, new_value):
        if DEBUG and self.name:
            print(
                "setting value for %s from %s to %s"
                % (self.name, self._value, new_value)
            )
        self._previous = self._value  # =old _value
        self._value = new_value
        self._dirty = False
        if not deep_eq(self._value, self._previous):
            if DEBUG and self.name:
                print("invalidating dependents for %s" % self.name)
            if self.on_change_handler:
                self.on_change_handler(self.value)
            self._invalidate_dependents()

    def recalc(self):
        # note: calls @value.setter!
        # tip: if the recalc callable is a method, the first arg (=first source.value) is used as 'self'
        new_value = self.recalc_handler(*[src.value for src in self.sources])
        if (
            new_value is not None
        ):  # None is never a valid cell value (eg, invalidated but no recalc function)
            self.value = new_value


CELL_CLASS_NAME = Cell.__name__  # 'Cell', see pype 'cells' property



def deep_eq(left, right):
    # both sides must be of the same type
    if type(left) != type(right):
        return False

    # pandas types
    if "pandas" in str(type(left)):
        return left.equals(right)

    # convert left and right to dicts if possible, skip if they can't be converted
    try:
        left = left.__dict__
        right = right.__dict__
    except Exception:
        pass

    # compare the two objects or dicts key by key
    if type(left) == dict:
        for key in left:
            # check if the key is present in the right dict, if not, we are not equals
            if key not in right:
                return False
            else:
                # compare the values if the key is present in both sides
                if not deep_eq(left[key], right[key]):
                    return False

        # check if any keys are present in right, but not in left
        for key in right:
            if key not in left:
                return False

        return True

    # check for each item in lists
    if type(left) == list:
        # right and left must have the same length
        if len(left) != len(right):
            return False

        # compare each item in the list
        for index in range(len(left)):
            if not deep_eq(left[index], right[index]):
                return False

    # do a standard comparison
    return left == right
